/*
 * USB Skeleton driver - 2.0
 *
 * Copyright (C) 2001-2004 Greg Kroah-Hartman (greg@kroah.com)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 * This driver is based on the 2.6.3 version of drivers/usb/usb-skeleton.c 
 * but has been rewritten to be easy to read and use, as no locks are now
 * needed anymore.
 *
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/kref.h>
#include <asm/uaccess.h>
#include <linux/usb.h>


/* Define these values to match your devices */
#define USB_ICU_VENDOR_ID	0x0abe
#define USB_ICU_PRODUCT_ID	0x0200

/* #define DEBUG 1 */

#ifdef DEBUG
static int debug = 1;
#define dbg_print(format, arg...) debug && printk(KERN_ERR "%s: " format, __FUNCTION__, ##arg)
#else
#define dbg_print(arg...)
#endif

/* table of devices that work with this driver */
static struct usb_device_id icu_table [] = {
	{ USB_DEVICE(USB_ICU_VENDOR_ID, USB_ICU_PRODUCT_ID) },
	{ }					/* Terminating entry */
};
MODULE_DEVICE_TABLE (usb, icu_table);


/* Get a minor range for your devices from the usb maintainer */
#define USB_ICU_MINOR_BASE	192

/* our private defines. if this grows any larger, use your own .h file */
#define MAX_TRANSFER		( PAGE_SIZE - 512 )
#define WRITES_IN_FLIGHT	8

/* Structure to hold all of our device specific stuff */
struct usb_icu {
	struct usb_device *	udev;			/* the usb device for this device */
	struct usb_interface *	interface;		/* the interface for this device */
	struct semaphore	limit_sem;		/* limiting the number of writes in progress */
	unsigned char *		bulk_in_buffer;		/* the buffer to receive data */
	size_t			bulk_in_size;		/* the size of the receive buffer */
	__u8			bulk_in_endpointAddr;	/* the address of the bulk in endpoint */
	__u8			bulk_out_endpointAddr;	/* the address of the bulk out endpoint */
	struct kref		kref;
};
#define to_icu_dev(d) container_of(d, struct usb_icu, kref)

static struct usb_driver icu_driver;

static void icu_delete(struct kref *kref)
{	
	struct usb_icu *dev = to_icu_dev(kref);

	usb_put_dev(dev->udev);
	kfree (dev->bulk_in_buffer);
	kfree (dev);
}

static int icu_open(struct inode *inode, struct file *file)
{
	struct usb_icu *dev;
	struct usb_interface *interface;
	int subminor;
	int retval = 0;

	dbg_print("opened\n");

	subminor = iminor(inode);

	interface = usb_find_interface(&icu_driver, subminor);
	if (!interface) {
		err ("%s - error, can't find device for minor %d",
		     __FUNCTION__, subminor);
		retval = -ENODEV;
		goto exit;
	}

	dev = usb_get_intfdata(interface);
	if (!dev) {
		retval = -ENODEV;
		goto exit;
	}

	/* increment our usage count for the device */
	kref_get(&dev->kref);

	/* save our object in the file's private structure */
	file->private_data = dev;

exit:
	return retval;
}

static int icu_release(struct inode *inode, struct file *file)
{
	struct usb_icu *dev;

	dbg_print("released\n");

	dev = (struct usb_icu *)file->private_data;
	if (dev == NULL)
		return -ENODEV;

	/* decrement the count on our device */
	kref_put(&dev->kref, icu_delete);
	return 0;
}

static ssize_t icu_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	struct usb_icu *dev;
	int retval = 0;
	int bytes_read;

	dev = (struct usb_icu *)file->private_data;
	
	dbg_print("ENTER, count = %d, pos = %d, bulk_in_size = %d\n", count, (int) *ppos, dev->bulk_in_size);

	/* do a blocking bulk read to get data from the device */

/**
 *      usb_bulk_msg - Builds a bulk urb, sends it off and waits for completion
 *      @usb_dev: pointer to the usb device to send the message to
 *      @pipe: endpoint "pipe" to send the message to
 *      @data: pointer to the data to send
 *      @len: length in bytes of the data to send
 *      @actual_length: pointer to a location to put the actual length transferred in bytes
 *      @timeout: time in msecs to wait for the message to complete before
 *              timing out (if 0 the wait is forever)
 */
	retval = usb_bulk_msg(dev->udev,
			      usb_rcvbulkpipe(dev->udev, dev->bulk_in_endpointAddr),
			      dev->bulk_in_buffer,
			      min(dev->bulk_in_size, count),
			      &bytes_read, 10000);

	/* if the read was successful, copy the data to userspace */
	if (!retval) {
		if (copy_to_user(buffer, dev->bulk_in_buffer, bytes_read))
			retval = -EFAULT;
		else
			retval = bytes_read;
	}
	dbg_print("EXIT, count = %d, bytes_read = %d\n", count, bytes_read);

	return retval;
}

static void icu_write_bulk_callback(struct urb *urb, struct pt_regs *regs)
{
	struct usb_icu *dev;

	dev = (struct usb_icu *)urb->context;

	/* sync/async unlink faults aren't errors */
	if (urb->status && 
	    !(urb->status == -ENOENT || 
	      urb->status == -ECONNRESET ||
	      urb->status == -ESHUTDOWN)) {
		dbg_print("%s - nonzero write bulk status received: %d",
		    __FUNCTION__, urb->status);
	}

	/* free up our allocated buffer */
	usb_buffer_free(urb->dev, urb->transfer_buffer_length, 
			urb->transfer_buffer, urb->transfer_dma);
	up(&dev->limit_sem);
}

static ssize_t icu_write(struct file *file, const char *user_buffer, size_t count, loff_t *ppos)
{
	struct usb_icu *dev;
	int retval = 0;
	struct urb *urb = NULL;
	char *buf = NULL;
	size_t writesize = min(count, (size_t)MAX_TRANSFER);

	dbg_print("ENTER, count = %d, pos = %d\n", count, (int) *ppos);

	dev = (struct usb_icu *)file->private_data;

	/* verify that we actually have some data to write */
	if (count == 0)
		goto exit;

	/* limit the number of URBs in flight to stop a user from using up all RAM */
	if (down_interruptible(&dev->limit_sem)) {
		retval = -ERESTARTSYS;
		goto exit;
	}

	/* create a urb, and a buffer for it, and copy the data to the urb */
	urb = usb_alloc_urb(0, GFP_KERNEL);
	if (!urb) {
		retval = -ENOMEM;
		goto error;
	}

	buf = usb_buffer_alloc(dev->udev, writesize, GFP_KERNEL, &urb->transfer_dma);
	if (!buf) {
		retval = -ENOMEM;
		goto error;
	}

	if (copy_from_user(buf, user_buffer, writesize)) {
		retval = -EFAULT;
		goto error;
	}

	/* initialize the urb properly */
	usb_fill_bulk_urb(urb, dev->udev,
			  usb_sndbulkpipe(dev->udev, dev->bulk_out_endpointAddr),
			  buf, writesize, icu_write_bulk_callback, dev);
	urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

	/* send the data out the bulk port */
	retval = usb_submit_urb(urb, GFP_KERNEL);
	if (retval) {
		err("%s - failed submitting write urb, error %d", __FUNCTION__, retval);
		goto error;
	}

	/* release our reference to this urb, the USB core will eventually free it entirely */
	usb_free_urb(urb);

exit:
	dbg_print("EXIT, writesize = %d\n", writesize);

	return writesize;

error:
	usb_buffer_free(dev->udev, writesize, buf, urb->transfer_dma);
	usb_free_urb(urb);
	up(&dev->limit_sem);
	return retval;
}

static struct file_operations icu_fops = {
	.owner =	THIS_MODULE,
	.read =		icu_read,
	.write =	icu_write,
	.open =		icu_open,
	.release =	icu_release,
};

/* 
 * usb class driver info in order to get a minor number from the usb core,
 * and to have the device registered with the driver core
 */
static struct usb_class_driver icu_class = {
	.name =		"icu%d",
	.fops =		&icu_fops,
	.minor_base =	USB_ICU_MINOR_BASE,
};

static int icu_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	struct usb_icu *dev = NULL;
	struct usb_host_interface *iface_desc;
	struct usb_endpoint_descriptor *endpoint;
	size_t buffer_size;
	int i;
	int retval = -ENOMEM;

	dbg_print("ENTER probe, id = %04x-%04x\n", id->idVendor, id->idProduct);

	/* allocate memory for our device state and initialize it */
	dev = kzalloc(sizeof(*dev), GFP_KERNEL);
	if (dev == NULL) {
		err("Out of memory");
		goto error;
	}
	kref_init(&dev->kref);
	sema_init(&dev->limit_sem, WRITES_IN_FLIGHT);

	dev->udev = usb_get_dev(interface_to_usbdev(interface));
	dev->interface = interface;

	/* set up the endpoint information */
	/* use only the first bulk-in and bulk-out endpoints */
	iface_desc = interface->cur_altsetting;
	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		endpoint = &iface_desc->endpoint[i].desc;

		if (!dev->bulk_in_endpointAddr &&
		    ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
					== USB_DIR_IN) &&
		    ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
					== USB_ENDPOINT_XFER_BULK)) {
			/* we found a bulk in endpoint */
			buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
			dev->bulk_in_size = buffer_size;
			dev->bulk_in_endpointAddr = endpoint->bEndpointAddress;
			dev->bulk_in_buffer = kmalloc(buffer_size, GFP_KERNEL);

			dbg_print("bulk_in: ep = %x, size = %d\n", endpoint->bEndpointAddress, buffer_size);

			if (!dev->bulk_in_buffer) {
				err("Could not allocate bulk_in_buffer");
				goto error;
			}
		}

		if (!dev->bulk_out_endpointAddr &&
		    ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
					== USB_DIR_OUT) &&
		    ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
					== USB_ENDPOINT_XFER_BULK)) {
			/* we found a bulk out endpoint */
			dev->bulk_out_endpointAddr = endpoint->bEndpointAddress;

			dbg_print("bulk_out: ep = %x\n", endpoint->bEndpointAddress);
		}
	}
	if (!(dev->bulk_in_endpointAddr && dev->bulk_out_endpointAddr)) {
		err("Could not find both bulk-in and bulk-out endpoints");
		goto error;
	}

	/* save our data pointer in this interface device */
	usb_set_intfdata(interface, dev);

	/* we can register the device now, as it is ready */
	retval = usb_register_dev(interface, &icu_class);
	if (retval) {
		/* something prevented us from registering this driver */
		err("Not able to get a minor for this device.");
		usb_set_intfdata(interface, NULL);
		goto error;
	}

	/* let the user know what node this device is now attached to */
	info("USB ICU device now attached to USBICU-%d", interface->minor);
	return 0;

error:
	if (dev)
		kref_put(&dev->kref, icu_delete);
	return retval;
}

static void icu_disconnect(struct usb_interface *interface)
{
	struct usb_icu *dev;
	int minor = interface->minor;

	/* prevent icu_open() from racing icu_disconnect() */
	lock_kernel();

	dev = usb_get_intfdata(interface);
	usb_set_intfdata(interface, NULL);

	/* give back our minor */
	usb_deregister_dev(interface, &icu_class);

	unlock_kernel();

	/* decrement our usage count */
	kref_put(&dev->kref, icu_delete);

	info("USB ICU #%d now disconnected", minor);
}

static struct usb_driver icu_driver = {
	.name =		"usbicu",
	.probe =	icu_probe,
	.disconnect =	icu_disconnect,
	.id_table =	icu_table,
};

static int __init usb_icu_init(void)
{
	int result;

	/* register this driver with the USB subsystem */
	printk("ICU host driver\n");
	result = usb_register(&icu_driver);
	if (result)
		err("usb_register failed. Error number %d", result);

	return result;
}

static void __exit usb_icu_exit(void)
{
	/* deregister this driver with the USB subsystem */
	usb_deregister(&icu_driver);
}

module_init (usb_icu_init);
module_exit (usb_icu_exit);

MODULE_DESCRIPTION("ICU usb host driver");
MODULE_LICENSE("GPL");
#ifdef DEBUG
module_param(debug, int, S_IRUGO|S_IWUSR);
MODULE_PARM_DESC(debug,"Debugenabled or not");
#endif
